package
{
	import away3d.containers.ObjectContainer3D;
	import away3d.containers.View3D;
	import away3d.core.base.Geometry;
	import away3d.core.base.SubGeometry;
	import away3d.core.base.SubMesh;
	import away3d.debug.AwayStats;
	import away3d.entities.Mesh;
	import away3d.lights.PointLight;
	import away3d.materials.ColorMaterial;
	import away3d.materials.TextureMaterial;
	import away3d.materials.lightpickers.StaticLightPicker;
	import away3d.primitives.Sphere;
	import away3d.textures.BitmapTexture;

	import flash.display.Sprite;
	import flash.events.Event;
	import flash.geom.Vector3D;

	/**
	 * SceneGraphDemo shows the use of SceneGraph objects such as Mesh, Light, Camera, ObjectContainer3D as well as
	 * the use of Geometries and SubGeometries/SubMeshes.
	 */

	[SWF(width="1024", height="576", frameRate="60", backgroundColor="0x000000")]
	public class SceneGraphDemo extends Sprite
	{
		[Embed(source="/../assets/earth_albedo.jpg")]
		private var EarthAlbedo : Class;

		[Embed(source="/../assets/earth_clouds.png")]
		private var EarthClouds : Class;

		[Embed(source="/../assets/earth_normals.jpg")]
		private var EarthNormals : Class;

		[Embed(source="/../assets/earth_specular.jpg")]
		private var EarthSpecular : Class;

		[Embed(source="/../assets/moon_albedo.jpg")]
		private var MoonAlbedo : Class;

		[Embed(source="/../assets/moon_normals.jpg")]
		private var MoonNormals : Class;

		[Embed(source="/../assets/stars.jpg")]
		private var Stars : Class;


		private var _view : View3D;

		private var _earthMoonContainer : ObjectContainer3D;
		private var _earth : Mesh;
		private var _moon : Sphere;
		private var _sun : Sphere;
		private var _skySphere : Sphere;
		private var _sunLight : PointLight;
		private var _cameraTargetPosition : Vector3D;

		private var _earthMaterial : TextureMaterial;
		private var _moonMaterial : TextureMaterial;
		private var _cloudsMaterial : TextureMaterial;

		private var _earthCycle : Number = 0;
		private var _moonCycle : Number = 0;


		public function SceneGraphDemo()
		{
			initView();

			initSkySphere();
			initSun();
			initEarthMoonMaterials();

			_earthMoonContainer = new ObjectContainer3D();
			_view.scene.addChild(_earthMoonContainer);

			initEarth();
			initMoon();
			addEventListener(Event.ENTER_FRAME, onEnterFrame);
		}

		/**
		 * Creates the view
		 */
		private function initView() : void
		{
			_view = new View3D();
			_view.camera.lens.far = 7000;
			_view.camera.y = 500;
			_view.camera.z = -1300;
			_cameraTargetPosition = new Vector3D();
			addChild(_view);
			addChild(new AwayStats(_view));
		}

		/**
		 * Creates the star sky sphere
		 */
		private function initSkySphere() : void
		{
			var skyTexture : BitmapTexture = new BitmapTexture(new Stars().bitmapData);
			var skyMaterial : TextureMaterial = new TextureMaterial(skyTexture);
			_skySphere = new Sphere(skyMaterial, 5000);

			// trick to turn the geometry inside out
			_skySphere.geometry.subGeometries[0].scale(-1);
			_view.scene.addChild(_skySphere);
		}

		/**
		 * Creates the materials for the earth and the moon
		 */
		private function initEarthMoonMaterials() : void
		{
			var lightPicker : StaticLightPicker;

			// init all textures
			var earthAlbedoTexture : BitmapTexture = new BitmapTexture(new EarthAlbedo().bitmapData);
			var earthCloudsTexture : BitmapTexture = new BitmapTexture(new EarthClouds().bitmapData);
			var moonAlbedoTexture : BitmapTexture = new BitmapTexture(new MoonAlbedo().bitmapData);

			_earthMaterial = new TextureMaterial(earthAlbedoTexture);
			_cloudsMaterial = new TextureMaterial(earthCloudsTexture);
			_moonMaterial = new TextureMaterial(moonAlbedoTexture);

			_earthMaterial.normalMap = new BitmapTexture(new EarthNormals().bitmapData);
			_earthMaterial.specularMap = new BitmapTexture(new EarthSpecular().bitmapData);
			_moonMaterial.normalMap = new BitmapTexture(new MoonNormals().bitmapData);

			lightPicker = new StaticLightPicker([ _sunLight ]);
			_earthMaterial.lightPicker = lightPicker;
			_cloudsMaterial.lightPicker = lightPicker;
			_moonMaterial.lightPicker = lightPicker;

			_earthMaterial.specular = .5;
			_moonMaterial.specular = .2;
			_cloudsMaterial.specular = .3;

			_cloudsMaterial.alphaBlending = true;
		}

		/**
		 * Creates the moon mesh
		 */
		private function initMoon() : void
		{
			_moon = new Sphere(_moonMaterial, 30);
			_earthMoonContainer.addChild(_moon);
		}

		/**
		 * Creates the sun mesh and material
		 */
		private function initSun() : void
		{
			var sunMaterial : ColorMaterial = new ColorMaterial(0xffffdd);

			_sun = new Sphere(sunMaterial, 90);
			_view.scene.addChild(_sun);

			_sunLight = new PointLight();
			_sunLight.ambient = 1;
			_sunLight.ambientColor = 0x505070;
			_sunLight.color = sunMaterial.color;
			_sun.addChild(_sunLight);	// we add the light as a child to the sun, so if we ever move the sun, we don't need to update the light
		}

		/**
		 * Creates an Earth with a Geometry containing two SubGeometry objects. As a result, the Mesh will have two
		 * SubMesh objects. This way, several materials can be used in one object without the overhead of two scene
		 * graph objects.
		 */
		private function initEarth() : void
		{
			// we won't actually be using the Sphere objects, but we'll be stealing their geometries to add to an Earth Mesh object
			var globe : Sphere = new Sphere(null, 85);
			var clouds : Sphere = new Sphere(null, 90);
			var earthSubGeom : SubGeometry;
			var cloudsSubGeom : SubGeometry;
			var subMesh : SubMesh;

			// create geometry for the earth itself
			var geometry : Geometry;

			// we'll use the Spheres' actual SubGeometries to build our own 2-sphere earth
			earthSubGeom = globe.geometry.subGeometries[0];
			cloudsSubGeom = clouds.geometry.subGeometries[0];

			// remove the SubGeometries
			globe.geometry.removeSubGeometry(earthSubGeom);
			clouds.geometry.removeSubGeometry(cloudsSubGeom);

			// create a Geometry from the SubGeometries we just stole
			geometry = new Geometry();
			geometry.addSubGeometry(earthSubGeom);
			geometry.addSubGeometry(cloudsSubGeom);

			_earth = new Mesh(_earthMaterial, geometry);	// use earthMaterial as default material

			// override the material for the clouds SubMesh
			subMesh = _earth.getSubMeshForSubGeometry(cloudsSubGeom);
			subMesh.material = _cloudsMaterial;

			_earthMoonContainer.addChild(_earth);

			// properly dispose of things we don't need anymore
			globe.dispose();
			clouds.dispose();
		}

		/**
		 * Update object positions and render
		 */
		private function onEnterFrame(event : Event) : void
		{
			// we're not exactly going for realism here :)

			// rotate the earth around its axis independently
			_earth.rotationY -= .5;

			// orbit earth and moon together around sun
			_earthMoonContainer.x = Math.sin(_earthCycle)*1000;
			_earthMoonContainer.z = Math.cos(_earthCycle)*800;
			_earthCycle += .01;

			// orbit moon independently around earth
			_moon.x = Math.cos(_moonCycle)*150;
			_moon.y = Math.sin(_moonCycle)*150;
			_moon.z = Math.sin(_moonCycle)*150;
			_moonCycle += .02;

			_view.render();

			// update camera target
			var earthScenePos : Vector3D = _earth.scenePosition;
			_cameraTargetPosition.x = earthScenePos.x*.5;
			_cameraTargetPosition.y = earthScenePos.y*.5;
			_cameraTargetPosition.z = earthScenePos.z*.5;
			_view.camera.lookAt(_cameraTargetPosition);
		}
	}
}
